// import {ResTab } from '../../../../res/res_mgr';
// import { RES_FILE } from '../../../../res/res_load';
// import { utf8Decode } from 'pi_sys/util/utf8';

// const FONT_SHEET: Map<number, FontCfg> = new Map();

export interface FontCfg {
	buffer: Uint8Array,
	index: Map<number, Glyph>,
	atlas: {
        type: string,
        distanceRange: number,
        size: number,
        width: number,
        height: number,
        yOrigin: string
    },
    metrics: {
        emSize: number,
        lineHeight: number,
        ascender: number,
        descender: number,
        underlineY: number,
        underlineThickness: number
    }
}

export interface FontAtlasData {
	name: string,
	atlas: {
        type: string,
        distanceRange: number,
        size: number,
        width: number,
        height: number,
        yOrigin: string
    },
    metrics: {
        emSize: number,
        lineHeight: number,
        ascender: number,
        descender: number,
        underlineY: number,
        underlineThickness: number
    },
	glyphs: Glyph[],
}

export interface Glyph {
    unicode: number,
    advance: number,
    planeBounds: {
        left: number,
        bottom: number,
        right: number,
        top: number
    },
    atlasBounds: {
        left: number,
        bottom: number,
        right: number,
        top: number
    }
}

export interface GlyphInfo {
    ox: number, //文字可见区域左上角相对于文字外边框的左上角在水平轴上的距离 百分比(实际百分比应该除以256，之所以这样，是为了压缩数据尺寸)
    oy: number, //文字可见区域左上角相对于文字外边框的左上角在垂直轴上的距离 百分比(实际百分比应该除以256，之所以这样，是为了压缩数据尺寸)
    width: number, // 纹理宽度
    height: number, // 纹理高度
    advance: number, // advance， 像素值
}

// export function addSdfFont(key: number, cfg: FontAtlasData, png: Uint8Array) {
// 	let map = new Map();
// 	for(let i = 0; i < cfg.glyphs.length; i++) {
// 		let glyph = cfg.glyphs[i];
// 		map.set(glyph.unicode, glyph);
// 	}

// 	FONT_SHEET.set(key, {
// 		atlas: cfg.atlas,
// 		metrics: cfg.metrics,
// 		buffer: png,
// 		index: map,
// 	});
// }

let sdfSuccessCallback, loadSdf;
export const setSdfSuccessCallback = (callback: (world: number, x: number,
	y: number,
	w: number,
	h: number,
	offset: Float32Array,
	data: Uint8Array) => void) => {
		sdfSuccessCallback = callback;
}

export const setSdfLoader = (loadFunc: (fontname: number, chars: Uint32Array) => Promise<any>) => {
	loadSdf = loadFunc;
}

export function drawSdf(world: number,font: number, chars: Uint32Array, info: Uint32Array, x: number, y: number, ww: number, hh: number) {
		if (loadSdf) {
			loadSdf(font, chars).then((r) => {
				// console.log(r);

				// const cfg = FONT_SHEET.get(font);
				// if (!cfg) {
				// 	console.log("font is not exist, font: ", font);
				// 	return;
				// }

				// let width = cfg.atlas.width;

				if (chars.length !== r.length) {
					console.error("load sdf char error");
					return;
				}

				// let pack = 4;
				let pack = 1;
				let buffer = new Uint8Array(ww * hh * pack);
				//
				for (let i = 0; i < chars.length; i++) {
					let bin = r[i];
					// let glyph = cfg.index.get(chars[i]);
					let info_index = i * 3;

					let o = info[info_index];
					let width = info[info_index + 1];
					let height = info[info_index + 2];

					let minWidth = Math.min(width, ww);
					let minHeight = Math.min(height, hh);
					// console.log("xxx===============", x);
					for (let i = 0; i < minHeight; i++) { // i代表行数
						for (let j = 0; j < minWidth; j++) { // 遍历一行中的每一列（j表列数）
							// 拷贝每个像素
							let y_line = i;
							let src_offset = (j + i * width) * pack;
							let dst_offset = (o + j + ww * y_line) * pack;
							// console.log(src_offset, dst_offset);
							buffer[dst_offset] = bin[src_offset];
							// buffer[dst_offset + 1] = bin[src_offset + 1];
							// buffer[dst_offset + 2] = bin[src_offset + 2];
							// buffer[dst_offset + 3] = bin[src_offset + 3];
							// console.log(i, j, dst_offset, src_offset);
						}
					}
				}
				
				if (sdfSuccessCallback) {
					// 异步回到，以免发生借用冲突
					setTimeout(() => {sdfSuccessCallback(world, x, y, ww, hh, buffer);}, 0);
					
				}
				
				// arr.push(resTab.load(RES_FILE, font_file_path, []), )
				// let font_file_path
				// resTab.load(RES_FILE, font_file_path, [])
				// // loadBinRes(resTab, url)
				// .then((res) => {
				// 	for (let i = 0; i < chars.length; i++) {
				// 		let x = info[i];

				// 	}
				// 	// let info = KTX.parse(res.link, false);
				// 	// (<any>window).__jsObj   = info.mipmaps[0].data; // new Uint8Array(res.link),
				// 	// (<any>window).__jsObj1  = hash;
				// 	// (<any>window).__jsObj.width = info.width;
				// 	// (<any>window).__jsObj.height = info.height;
				// 	// modifySize((<any>window).__jsObj, width, height);
				// 	// 这里可能需要压缩纹理文件后缀表明是否透明，暂时认为全是透明纹理 TODO
				// 	callback(1, info.format, 0, hash, info.width, info.height, info.mipmaps[0].data, info.mipmaps[0].data.length);
			
				// }).catch((e) => {
				// 	console.log("load font fail, ", e);
				// });

			});
		}

		

}

export class AtomTable{
	stringMap /*Map<number, string>*/ = new Map();
	numberMap /*Map<string, number>*/ = new Map();

	get_string(k) {
		return this.stringMap.get(k)
	}

	get_number(k) {
		return this.numberMap.get(k)
	}

	set(n, s) {
		this.stringMap.set(n, s);
		this.numberMap.set(s, n);
	}
}

export const atomTable = new AtomTable();

export function fillBackGround(canvas, ctx, x, y) {
	window["canvas11"] = canvas;
	if (x !== canvas.width) {
		canvas.width = x;
	}
	if (y !== canvas.height) {
		canvas.height = y;
	}
	// ctx.fillStyle = "#00f";
	ctx.fillStyle = "#f0f";
	ctx.fillRect(0, 0, canvas.width, canvas.height);
}

export function drawCharWithStroke(ctx, ch_code, x, y) {
	var ch = String.fromCharCode(ch_code);
	//fillText 和 strokeText 的顺序对最终效果会有影响， 为了与css text-stroke保持一致， 应该fillText在前
	ctx.strokeText(ch, x, y);
	ctx.fillText(ch, x, y);
}

export function drawChar(ctx, ch_code, x, y) {
	var ch = String.fromCharCode(ch_code);
	//fillText 和 strokeText 的顺序对最终效果会有影响， 为了与css text-stroke保持一致， 应该fillText在前
    ctx.fillText(ch, x, y);
}

export function setFont(ctx, weight, fontSize, font/*number */, strokeWidth) {
	font = atomTable.get_string(font);
	var weight;
	if (weight <= 300 ) {
		weight = "lighter";
	} else if (weight < 700 ) {
		weight = "normal";
	} else if (weight < 900 ) {
		weight = "bold";
	} else {
		weight = "bolder";
	}
	ctx.font = weight + " " + fontSize + "px " + font;
	ctx.fillStyle = "#0f0";
	ctx.textBaseline = "top";

	if(strokeWidth > 0) {
		ctx.lineWidth = strokeWidth;
		ctx.strokeStyle = "#f00";
	}
}

export function useVao() {
	return false;
	var u = navigator.userAgent.toLowerCase(); 
	return u.indexOf("ipad") < 0 && u.indexOf("iphone") < 0;
}

export function measureText(ctx, ch, font_size, name/**number*/) {
	name = atomTable.get_string(name);
	ctx.font = font_size + "px " + name;
	return ctx.measureText(String.fromCharCode(ch)).width;
}

export function loadImage(image_name, callback) {
	var image = new Image();
	image.onload = function() {
		callback(image_name.endsWith("png")?1:0, -1, 0, image_name, image.width, image.height, image);
	};
	image.src = image_name;
}